// ----------------------------------------------------------------------------
//
// Matrix multiplication (multi-threaded client/server version).
// (c) Wolfgang Schreiner, 2017.
//
// ----------------------------------------------------------------------------

import java.io.*;
import java.net.*;

public class MatMultNet
{
  // name and port of server
  private final static String URL = "localhost";
  private final static int port = 9999;

  // matrix dimension
  private static int N; 

  // number of threads
  private static int T;
 
  // shared data used by threads
  private static double[][] A;
  private static double[][] B;
  private static double[][] C;

  // -------------------------------------------------------------------------
  // 
  // a thread in charge of filling C[begin..end-1]
  //
  // -------------------------------------------------------------------------
  private static final class MultThread extends Thread
  {
    private int begin;
    private int end;
    public MultThread(int begin, int end)
    {
      this.begin = begin;
      this.end = end;
    }
    public void run()
    {
      for (int i = begin; i < end; i++) 
      {
        for (int j = 0; j < N; j++) 
        {
	      C[i][j] = 0;
	      for (int k = 0; k < N; k++) 
	        C[i][j] += A[i][k]*B[k][j];
        }
      }
    }
  }

  // -------------------------------------------------------------------------
  // 
  // multiply A by B giving C
  // 
  // -------------------------------------------------------------------------
  private static void multiply() 
  {
    int n = N/T;
    
    Thread[] thread = new MultThread[T];
    for (int i = 0; i < T; i++) 
    {
	  thread[i] = new MultThread(i*n, Math.min((i+1)*n,N));
      thread[i].start();
    }

    try
    {
      for (int i = 0; i < T; i++) 
        thread[i].join();
    }
    catch(InterruptedException e) { }
  }

  // -------------------------------------------------------------------------
  //
  // main program
  //
  // --------------------------------------------------------------------------
  public static void main(String[] args) 
  {
    if (args.length != 1)
    {
      System.out.println("Usage: MatMultThreads -client | -server");
      return;
    }
    if (args[0].equals("-client"))
      client();
    else 
      server();
  }
  
  // ----------------------------------------------------------------------------
  //
  // the client
  //
  // ----------------------------------------------------------------------------
  static void client()
  {
    try
    {
      BufferedReader console = 
        new BufferedReader(new InputStreamReader(System.in));
      while (true)
      {
        String line = console.readLine();
        if (line == null) return;   
        Socket socket = new Socket(URL, port);
        BufferedReader in = 
          new BufferedReader(new InputStreamReader(socket.getInputStream()));
        PrintWriter out = 
          new PrintWriter(new OutputStreamWriter(socket.getOutputStream()), true);
        out.println(line);
        String answer = in.readLine();
        if (answer == null) return;
        System.out.println(answer);
      }
    }
    catch(IOException e)
    {
      System.out.println("Exception: " + e);
      System.exit(-1);
    }
  }

  
  // ----------------------------------------------------------------------------
  //
  // the server
  //
  // ----------------------------------------------------------------------------
  public static void server() 
  {
    try
    {
      ServerSocket server = new ServerSocket(port);
      while (true)
      {
        Socket socket = server.accept();
        BufferedReader in = 
          new BufferedReader(new InputStreamReader(socket.getInputStream()));
        PrintWriter out = 
          new PrintWriter(new OutputStreamWriter(socket.getOutputStream()), true);
        String line = in.readLine();
        if (line == null) return;
        String[] args = line.split(" ");
        if (args.length != 2)
        {
          out.println("Usage: MatMultNet <N> <T>");
          continue;
        }
        try
        {
          N = Integer.parseInt(args[0]);
          T = Integer.parseInt(args[1]);
        }
        catch(NumberFormatException e)
        {
          out.println("Invalid argument (number expected): " + args[0]);
          continue;
        } 
        A = new double[N][N];
        B = new double[N][N];
        C = new double[N][N];
        long t1 = System.currentTimeMillis();
        multiply();
        long t2 = System.currentTimeMillis();
        out.println((t2-t1) + " ms");
      }
    }
    catch(IOException e)
    {
      System.out.println("Exception: " + e);
      System.exit(-1);
    }

  }
}

